package com.example.rgbblecontrol;

import android.Manifest;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGatt;
import android.bluetooth.BluetoothGattCallback;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.bluetooth.BluetoothManager;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.os.Bundle;
import android.os.Handler;
import android.util.Log;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.ListView;
import android.widget.Toast;

import androidx.annotation.NonNull;
import androidx.appcompat.app.AppCompatActivity;
import androidx.core.app.ActivityCompat;

import com.skydoves.colorpickerview.ColorEnvelope;
import com.skydoves.colorpickerview.ColorPickerView;
import com.skydoves.colorpickerview.listeners.ColorEnvelopeListener;

import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

/**
 * App that connects to a microcontroller (eg. Arduino, Raspberry PI Pico W) to control an RGB Led.
 * It will send the selected color to the microcontroller, which changes the Led color accordingly.
 * More information:
 * https://home.et.utwente.nl/slootenvanf/2025/03/07/create-rgb-led-controller-app/
 */
public class MainActivity extends AppCompatActivity {

    private static final int REQUEST_ENABLE_BT = 1;
    private static final int REQUEST_PERMISSION_LOCATION = 2;
    private static final String TAG = "MainActivity";

    private BluetoothAdapter bluetoothAdapter;
    private BluetoothGatt bluetoothGatt;
    private BluetoothGattCharacteristic writeCharacteristic;

    private List<BluetoothDevice> bluetoothDevices = new ArrayList<>();
    private ArrayAdapter<String> deviceArrayAdapter;

    // UUIDs for the UART service (Generic Attribute Profile (GATT) and characteristic)
    private static final UUID UART_SERVICE_UUID = UUID.fromString("6E400001-B5A3-F393-E0A9-E50E24DCCA9E");
    private static final UUID TX_CHARACTERISTIC_UUID = UUID.fromString("6E400002-B5A3-F393-E0A9-E50E24DCCA9E");
    private static final UUID RX_CHARACTERISTIC_UUID = UUID.fromString("6E400003-B5A3-F393-E0A9-E50E24DCCA9E");

    private int lastColor = 0;
    private Handler handler = new Handler();
    private Runnable sendColorRunnable;

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);

        BluetoothManager bluetoothManager = (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        bluetoothAdapter = bluetoothManager.getAdapter();

        Button scanButton = findViewById(R.id.scanButton);
        ListView deviceList = findViewById(R.id.deviceList);
        ColorPickerView colorPickerView = findViewById(R.id.colorPickerView);

        deviceArrayAdapter = new ArrayAdapter<>(this, android.R.layout.simple_list_item_1);
        deviceList.setAdapter(deviceArrayAdapter);

        scanButton.setOnClickListener(v -> {
            if (bluetoothAdapter == null || !bluetoothAdapter.isEnabled()) {
                Toast.makeText(MainActivity.this, "Please enable Bluetooth", Toast.LENGTH_SHORT).show();
            } else {
                if (ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED ||
                        ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.BLUETOOTH_SCAN) != PackageManager.PERMISSION_GRANTED ||
                        ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.BLUETOOTH_CONNECT) != PackageManager.PERMISSION_GRANTED) {
                    ActivityCompat.requestPermissions(MainActivity.this, new String[]{
                            Manifest.permission.ACCESS_FINE_LOCATION,
                            Manifest.permission.BLUETOOTH_SCAN,
                            Manifest.permission.BLUETOOTH_CONNECT
                    }, REQUEST_PERMISSION_LOCATION);
                } else {
                    scanForDevices();
                }
            }
        });

        deviceList.setOnItemClickListener((parent, view, position, id) -> {
            BluetoothDevice device = bluetoothDevices.get(position);
            connectToDevice(device);
        });

        colorPickerView.setColorListener((ColorEnvelopeListener) (envelope, fromUser) -> {
            int color = envelope.getColor();
            if (Math.abs(color - lastColor) > 5000) { // Adjust the threshold as needed
                lastColor = color;
                if (sendColorRunnable != null) {
                    handler.removeCallbacks(sendColorRunnable);
                }
                sendColorRunnable = () -> {
                    if (bluetoothGatt != null && writeCharacteristic != null) {
                        if (bluetoothGatt.connect()) {
                            String colorString = envelope.getArgb()[1] + "," + envelope.getArgb()[2] + "," + envelope.getArgb()[3];
                            writeCharacteristic.setValue(colorString);
                            bluetoothGatt.writeCharacteristic(writeCharacteristic);
                            Toast.makeText(MainActivity.this, "Color sent to device: " + colorString, Toast.LENGTH_SHORT).show();
                        } else {
                            Toast.makeText(MainActivity.this, "Failed to send color: BLE connection lost", Toast.LENGTH_SHORT).show();
                        }
                    } else {
                        Toast.makeText(MainActivity.this, "Failed to send color: No connection", Toast.LENGTH_SHORT).show();
                    }
                };
                handler.postDelayed(sendColorRunnable, 250); // Adjust the delay as needed
            }
        });

        scanForDevices();
    } // end onCreate() method

    private void scanForDevices() {
        bluetoothDevices.clear();
        deviceArrayAdapter.clear();
        if (ActivityCompat.checkSelfPermission(this, Manifest.permission.BLUETOOTH_SCAN) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.BLUETOOTH_SCAN}, REQUEST_PERMISSION_LOCATION);
            return;
        }
        bluetoothAdapter.startDiscovery();
        registerReceiver(new BluetoothDeviceReceiver(), new IntentFilter(BluetoothDevice.ACTION_FOUND));
        Toast.makeText(this, "Scanning for devices...", Toast.LENGTH_SHORT).show();
    }

    private void connectToDevice(BluetoothDevice device) {
        if (ActivityCompat.checkSelfPermission(this, Manifest.permission.BLUETOOTH_CONNECT) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.BLUETOOTH_CONNECT}, REQUEST_PERMISSION_LOCATION);
            return;
        }
        try {
            bluetoothGatt = device.connectGatt(this, false, new BluetoothGattCallback() {
                @Override
                public void onConnectionStateChange(@NonNull BluetoothGatt gatt, int status, int newState) {
                    if (newState == BluetoothGatt.STATE_CONNECTED) {
                        runOnUiThread(() -> Toast.makeText(MainActivity.this, "Connected to device", Toast.LENGTH_SHORT).show());
                        if (ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.BLUETOOTH_CONNECT) != PackageManager.PERMISSION_GRANTED) {
                            ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.BLUETOOTH_CONNECT}, REQUEST_PERMISSION_LOCATION);
                            return;
                        }
                        gatt.discoverServices();
                    } else if (newState == BluetoothGatt.STATE_DISCONNECTED) {
                        runOnUiThread(() -> Toast.makeText(MainActivity.this, "Disconnected from device", Toast.LENGTH_SHORT).show());
                        bluetoothGatt.close();
                    }
                }

                @Override
                public void onServicesDiscovered(@NonNull BluetoothGatt gatt, int status) {
                    if (status == BluetoothGatt.GATT_SUCCESS) {
                        runOnUiThread(() -> Toast.makeText(MainActivity.this, "Services discovered", Toast.LENGTH_SHORT).show());
                        if (ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.BLUETOOTH_CONNECT) != PackageManager.PERMISSION_GRANTED) {
                            ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.BLUETOOTH_CONNECT}, REQUEST_PERMISSION_LOCATION);
                            return;
                        }
                        BluetoothGattService service = gatt.getService(UART_SERVICE_UUID);
                        if (service != null) {
                            writeCharacteristic = service.getCharacteristic(TX_CHARACTERISTIC_UUID);
                            runOnUiThread(() -> Toast.makeText(MainActivity.this, "Characteristic found", Toast.LENGTH_SHORT).show());
                        } else {
                            runOnUiThread(() -> Toast.makeText(MainActivity.this, "Service not found", Toast.LENGTH_SHORT).show());
                        }
                    } else {
                        runOnUiThread(() -> Toast.makeText(MainActivity.this, "Failed to discover services", Toast.LENGTH_SHORT).show());
                    }
                }
            });
        } catch (SecurityException e) {
            e.printStackTrace();
            Toast.makeText(this, "Bluetooth connection failed due to missing permissions", Toast.LENGTH_SHORT).show();
        }
    }

    @Override
    public void onRequestPermissionsResult(int requestCode, @NonNull String[] permissions, @NonNull int[] grantResults) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults);
        if (requestCode == REQUEST_PERMISSION_LOCATION) {
            if (grantResults.length > 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                scanForDevices();
            } else {
                Toast.makeText(this, "Location and Bluetooth permissions are required to scan for BLE devices", Toast.LENGTH_SHORT).show();
            }
        }
    }

    private class BluetoothDeviceReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            if (BluetoothDevice.ACTION_FOUND.equals(action)) {
                BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                if (device != null) {
                    if (ActivityCompat.checkSelfPermission(MainActivity.this, Manifest.permission.BLUETOOTH_CONNECT) != PackageManager.PERMISSION_GRANTED) {
                        ActivityCompat.requestPermissions(MainActivity.this, new String[]{Manifest.permission.BLUETOOTH_CONNECT}, REQUEST_PERMISSION_LOCATION);
                        return;
                    }
                    String deviceName = device.getName() != null ? device.getName() : "Unknown Device";
                    bluetoothDevices.add(device);
                    deviceArrayAdapter.add(deviceName + "\n" + device.getAddress());
                    deviceArrayAdapter.notifyDataSetChanged();
                }
            }
        }
    }
}